@@ -28,7 +28,7 @@ import sys
 import stat
 import time
 import logging
-from signal import alarm, signal, SIGALRM, SIGINT, SIGKILL
+from signal import alarm, signal, SIGALRM, SIGINT, SIGKILL, SIGHUP
 import subprocess
 import fcntl
 import random
@@ -43,6 +43,7 @@ MAX_WAIT_TEST_RUN = 45*60
 
 interrupted = False
 old_sigint_handler = None
+got_sighup = False
 
 def setup_logging(log_level, log_file_name):
     logger = logging.getLogger()
@@ -77,7 +78,9 @@ class Config(piupartslib.conf.Config):
                 "master-command": None,
                 "log-file": "piuparts-master.log",
                 "mirror": None,
-                "piuparts-cmd": "sudo piuparts",
+                "piuparts-command": "sudo piuparts",
+                "piuparts-flags": "",
+                "tmpdir": "/tmp/piuparts",
                 "distro": None,
                 "area": None,
                 "chroot-tgz": None,
@@ -105,6 +108,11 @@ def sigint_handler(signum, frame):
     print 'Press Ctrl-C again to abort now.'
     signal(SIGINT, old_sigint_handler)
 
+def sighup_handler(signum, frame):
+    global got_sighup
+    got_sighup = True
+    print 'SIGHUP: Will flush finished logs.'
+
 
 class MasterIsBusy(Exception):
 
@@ -124,12 +132,24 @@ class MasterDidNotGreet(Exception):
         self.args = "Master did not start with 'hello'"
 
 
+class MasterCommunicationFailed(Exception):
+
+    def __init__(self):
+        self.args = "Communication with master failed"
+
+
 class MasterIsCrazy(Exception):
 
     def __init__(self):
         self.args = "Master said something unexpected"
 
 
+class MasterCantRecycle(Exception):
+
+    def __init__(self):
+        self.args = "Master has nothing to recycle"
+
+
 class Slave:
 
     def __init__(self):
@@ -141,15 +161,21 @@ class Slave:
         self._master_command = None
 
     def _readline(self):
-        line = self._from_master.readline()
+        try:
+            line = self._from_master.readline()
+        except IOError:
+            raise MasterCommunicationFailed()
         logging.debug("<< " + line.rstrip())
         return line
 
     def _writeline(self, *words):
         line = " ".join(words)
         logging.debug(">> " + line)
-        self._to_master.write(line + "\n")
-        self._to_master.flush()
+        try:
+            self._to_master.write(line + "\n")
+            self._to_master.flush()
+        except IOError:
+            raise MasterCommunicationFailed()
 
     def set_master_host(self, host):
         logging.debug("Setting master host to %s" % host)
@@ -189,8 +215,10 @@ class Slave:
 
     def close(self):
         logging.debug("Closing connection to master")
-        self._from_master.close()
-        self._to_master.close()
+        if self._from_master is not None:
+            self._from_master.close()
+        if self._to_master is not None:
+            self._to_master.close()
         self._from_master = self._to_master = None
         logging.info("Connection to master closed")
 
@@ -220,6 +248,22 @@ class Slave:
         else:
             raise MasterIsCrazy()
 
+    def enable_recycling(self):
+        self._writeline("recycle")
+        line = self._readline()
+        words = line.split()
+        if line != "ok\n":
+            raise MasterCantRecycle()
+
+    def get_idle(self):
+        self._writeline("idle")
+        line = self._readline()
+        words = line.split()
+        if words and words[0] == "ok":
+            return int(words[1])
+        else:
+            raise MasterIsCrazy()
+
     def reserve(self):
         self._writeline("reserve")
         line = self._readline()
@@ -234,6 +278,16 @@ class Slave:
         else:
             raise MasterIsCrazy()
 
+    def unreserve(self, filename):
+        basename = os.path.basename(filename)
+        package, rest = basename.split("_", 1)
+        version = rest[:-len(".log")]
+        logging.info("Unreserve: %s %s" % (package, version))
+        self._writeline("unreserve", package, version)
+        line = self._readline()
+        if line != "ok\n":
+            raise MasterNotOK()
+
     def _reserved_filename(self, name, version):
         return os.path.join("reserved",  "%s_%s.log" % (name, version))
 
@@ -262,6 +316,7 @@ class Section:
         self._config.read(CONFIG_FILE)
         self._error_wait_until = 0
         self._idle_wait_until = 0
+        self._recycle_wait_until = 0
         self._slave_directory = os.path.abspath(self._config["slave-directory"])
         if not os.path.exists(self._slave_directory):
             os.makedirs(self._slave_directory)
@@ -283,12 +338,14 @@ class Section:
 
         self._slave = Slave()
 
-    def _connect_to_master(self):
+    def _connect_to_master(self, recycle=False):
         self._slave.set_master_host(self._config["master-host"])
         self._slave.set_master_user(self._config["master-user"])
         self._slave.set_master_directory(self._config["master-directory"])
         self._slave.set_master_command(self._config["master-command"] + " " + self._config.section)
         self._slave.connect_to_master(self._config["log-file"])
+        if recycle:
+            self._slave.enable_recycling()
 
     def _check_tarball(self):
         oldcwd = os.getcwd()
@@ -306,18 +363,38 @@ class Section:
 
         os.chdir(oldcwd)
 
+    def _count_submittable_logs(self):
+        files = 0
+        for logdir in ["pass", "fail", "untestable"]:
+            for basename in os.listdir(os.path.join(self._slave_directory, logdir)):
+                if basename.endswith(".log"):
+                    files += 1
+        return files
+
     def precedence(self):
         return int(self._config["precedence"])
 
-    def sleep_until(self):
+    def sleep_until(self, recycle=False):
+        if recycle:
+            return max(self._error_wait_until, self._recycle_wait_until)
         return max(self._error_wait_until, self._idle_wait_until)
 
-    def run(self):
-        if time.time() < self.sleep_until():
+
+    def run(self, do_processing=True, recycle=False):
+        if time.time() < self.sleep_until(recycle=recycle):
+            return 0
+
+        if not do_processing and self._count_submittable_logs() == 0:
             return 0
 
         logging.info("-------------------------------------------")
-        logging.info("Running section %s (precedence=%d)" % (self._config.section, self.precedence()))
+        action = "Running"
+        if recycle:
+            action = "Recycling"
+        if not do_processing:
+            action = "Flushing"
+        logging.info("%s section %s (precedence=%d)" \
+                     % (action, self._config.section, self.precedence()))
         self._config = Config(section=self._config.section, defaults_section="global")
         self._config.read(CONFIG_FILE)
 
@@ -332,57 +409,96 @@ class Section:
             self._error_wait_until = time.time() + 3600
             return 0
 
-        lock = open(os.path.join(self._slave_directory, "slave.lock"), "we")
+        with open(os.path.join(self._slave_directory, "slave.lock"), "we") as lock:
+            oldcwd = os.getcwd()
+            os.chdir(self._slave_directory)
+            try:
+                fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
+            except IOError:
+                logging.info("busy")
+                self._error_wait_until = time.time() + 900
+            else:
+                if self._talk_to_master(fetch=do_processing, recycle=recycle):
+                    if do_processing:
+                        if not self._slave.get_reserved():
+                            self._idle_wait_until = time.time() + int(self._config["idle-sleep"])
+                            if recycle:
+                                self._recycle_wait_until = self._idle_wait_until + 3600
+                        else:
+                            processed = self._process()
+                            # put this section at the end of the round-robin runnable queue
+                            self._idle_wait_until = time.time()
+                            self._recycle_wait_until = time.time()
+                            return processed
+            finally:
+                os.chdir(oldcwd)
+        return 0
+
+
+    def _talk_to_master(self, fetch=False, unreserve=False, recycle=False):
+        flush = self._count_submittable_logs() > 0
+        fetch = fetch and not self._slave.get_reserved()
+        if not flush and not fetch:
+            return True
+
         try:
-            fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
-        except IOError:
-            logging.info("busy")
-            self._error_wait_until = time.time() + 900
-            lock.close()
-            return 0
-
-        oldcwd = os.getcwd()
-        os.chdir(self._slave_directory)
-
-        ret = self._run()
-
-        os.chdir(oldcwd)
-        lock.close()
-        return ret
-
-    def _run(self):
-        try:
-            self._connect_to_master()
+            self._connect_to_master(recycle=recycle)
         except KeyboardInterrupt:
             raise
         except MasterIsBusy:
             logging.error("master is busy")
             self._error_wait_until = time.time() + random.randrange(60, 180)
-            return 0
-        except:
+        except MasterCantRecycle:
+            logging.error("master has nothing to recycle")
+            self._recycle_wait_until = max(time.time(), self._idle_wait_until) + 3600
+        except (MasterDidNotGreet, MasterIsCrazy, MasterCommunicationFailed):
             logging.error("connection to master failed")
             self._error_wait_until = time.time() + 900
-            return 0
+        else:
+            try:
+                for logdir in ["pass", "fail", "untestable"]:
+                    for basename in os.listdir(logdir):
+                        if basename.endswith(".log"):
+                            fullname = os.path.join(logdir, basename)
+                            self._slave.send_log(self._config.section, logdir, fullname)
+                            os.remove(fullname)
 
-        for logdir in ["pass", "fail", "untestable"]:
-            for basename in os.listdir(logdir):
-                if basename.endswith(".log"):
-                    fullname = os.path.join(logdir, basename)
-                    self._slave.send_log(self._config.section, logdir, fullname)
-                    os.remove(fullname)
+                if unreserve:
+                    for logdir in ["new", "reserved"]:
+                        for basename in os.listdir(logdir):
+                            if basename.endswith(".log"):
+                                fullname = os.path.join(logdir, basename)
+                                self._slave.unreserve(fullname)
+                                os.remove(fullname)
 
-        if not self._slave.get_reserved():
-            max_reserved = int(self._config["max-reserved"])
-            while len(self._slave.get_reserved()) < max_reserved and self._slave.reserve():
-                pass
+                if fetch:
+                    max_reserved = int(self._config["max-reserved"])
+                    idle = self._slave.get_idle()
+                    if idle > 0:
+                        idle = min(idle, int(self._config["idle-sleep"]))
+                        logging.info("idle (%d)" % idle)
+                        if not recycle:
+                            self._idle_wait_until = time.time() + idle
+                        else:
+                            self._recycle_wait_until = time.time() + idle
+                        return 0
+                    while len(self._slave.get_reserved()) < max_reserved and self._slave.reserve():
+                        pass
+                    self._slave.get_status(self._config.section)
+            except MasterNotOK:
+                logging.error("master did not respond with 'ok'")
+                self._error_wait_until = time.time() + 900
+            except (MasterIsCrazy, MasterCommunicationFailed):
+                logging.error("communication with master failed")
+                self._error_wait_until = time.time() + 900
+            else:
+                return True
+        finally:
+            self._slave.close()
+        return False
 
-        self._slave.get_status(self._config.section)
-        self._slave.close()
-
-        if not self._slave.get_reserved():
-            self._idle_wait_until = time.time() + int(self._config["idle-sleep"])
-            return 0
 
+    def _process(self):
         if self._config["distro"]:
             distros = [self._config["distro"]]
         else:
@@ -408,6 +524,8 @@ class Section:
         test_count = 0
         self._check_tarball()
         for package_name, version in self._slave.get_reserved():
+            if got_sighup:
+                break
             test_count += 1
             if package_name in packages_file:
                 package = packages_file[package_name]
@@ -426,7 +544,10 @@ class Section:
                             "Package %s not found\n" % package_name)
             self._slave.forget_reserved(package_name, version)
             if interrupted:
-                raise KeyboardInterrupt
+                break
+        self._talk_to_master(unreserve=interrupted)
+        if interrupted:
+            raise KeyboardInterrupt
         return test_count
 
 
@@ -504,7 +625,7 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
     except KeyboardInterrupt:
         print '\nSlave interrupted by the user, cleaning up...'
         try:
-            kill_subprocess(p, kill_all)
+            terminate_subprocess(p, kill_all)
         except KeyboardInterrupt:
             print '\nTerminating piuparts was interrupted... manual cleanup still neccessary.'
             raise
@@ -516,9 +637,12 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
 def test_package(config, package, packages_files):
     global old_sigint_handler
     old_sigint_handler = signal(SIGINT, sigint_handler)
-    logging.info("Testing package %s/%s %s" % (config.section, package["Package"], package["Version"]))
 
-    output_name = log_name(package["Package"], package["Version"])
+    pname = package["Package"]
+    pvers = package["Version"]
+    logging.info("Testing package %s/%s %s" % (config.section, pname, pvers))
+
+    output_name = log_name(pname, pvers)
     logging.debug("Opening log file %s" % output_name)
     new_name = os.path.join("new", output_name)
     output = file(new_name, "we")
@@ -528,41 +652,57 @@ def test_package(config, package, packages_files):
     package.dump(output)
     output.write("\n")
 
-    base_command = config["piuparts-cmd"].split()
+    base_command = config["piuparts-command"].split()
+    if config["piuparts-flags"]:
+        base_command.extend(config["piuparts-flags"].split())
     if config["mirror"]:
         base_command.extend(["--mirror", config["mirror"]])
+    if config["tmpdir"]:
+        base_command.extend(["--tmpdir", config["tmpdir"]])
 
     ret = 0
     if config["chroot-tgz"]:
         command = base_command[:]
         command.extend(["-b", config["chroot-tgz"]])
         command.extend(["-d", config["distro"]])
+        command.append("--no-upgrade-test")
         if config["keep-sources-list"] in ["yes", "true"]:
             command.append("--keep-sources-list")
-        command.extend(["--apt", "%s=%s" % (package["Package"], package["Version"])])
+        command.extend(["--apt", "%s=%s" % (pname, pvers)])
 
         output.write("Executing: %s\n" % " ".join(command))
         ret,f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN)
+        if not f or f[-1] != '\n':
+            f += '\n'
+        output.write(f)
+        lastline = f.split('\n')[-2]
         if ret < 0:
-            output.write(f + "\n *** Process KILLED - exceed maximum run time ***\n")
-        else:
-            output.write(f)
+            output.write(" *** Process KILLED - exceed maximum run time ***\n")
+        elif not "piuparts run ends" in lastline:
+            ret += 1024
+            output.write(" *** PIUPARTS OUTPUT INCOMPLETE ***\n");
 
     if ret == 0 and config["upgrade-test-chroot-tgz"] and upgrade_testable(config, package, packages_files):
         command = base_command[:]
         command.extend(["-b", config["upgrade-test-chroot-tgz"]])
         for distro in config["upgrade-test-distros"].split():
             command.extend(["-d", distro])
-        command.extend(["--apt", "%s=%s" % (package["Package"], package["Version"])])
+        command.extend(["--apt", "%s=%s" % (pname, pvers)])
 
         output.write("Executing: %s\n" % " ".join(command))
         ret,f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN)
+        if not f or f[-1] != '\n':
+            f += '\n'
+        output.write(f)
+        lastline = f.split('\n')[-2]
         if ret < 0:
             output.write(" *** Process KILLED - exceed maximum run time ***\n")
-        else:
-            output.write(f)
+        elif not "piuparts run ends" in lastline:
+            ret += 1024
+            output.write(" *** PIUPARTS OUTPUT INCOMPLETE ***\n");
 
     output.write("\n")
+    output.write("ret=%d\n" % ret)
     output.write(time.strftime("End: %Y-%m-%d %H:%M:%S %Z\n",
                                time.gmtime()))
     output.close()
@@ -579,18 +719,26 @@ def create_chroot(config, tarball, distro):
     output_name = tarball + ".log"
     logging.debug("Opening log file %s" % output_name)
     logging.info("Creating new tarball %s" % tarball)
-    command = "%s -ad %s -s %s.new -m %s dpkg" % \
-                (config["piuparts-cmd"], distro, tarball, config["mirror"])
+    command = config["piuparts-command"].split()
+    if config["piuparts-flags"]:
+        command.extend(config["piuparts-flags"].split())
+    if config["mirror"]:
+        command.extend(["--mirror", config["mirror"]])
+    if config["tmpdir"]:
+        command.extend(["--tmpdir", config["tmpdir"]])
+    command.extend(["-d", distro])
+    command.extend(["-s", tarball + ".new"])
+    command.extend(["--apt", "dpkg"])
     output = file(output_name, "w")
     output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n\n",
                                time.gmtime()))
-    output.write("Executing: " + command)
-    logging.debug("Executing: " + command)
-    f = os.popen("{ %s; } 2>&1" % command, "r")
-    for line in f:
+    output.write("Executing: " + " ".join(command) + "\n\n")
+    logging.debug("Executing: " + " ".join(command))
+    p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+    for line in p.stdout:
         output.write(line)
         logging.debug(">> " + line.rstrip())
-    f.close()
+    p.wait()
     output.write(time.strftime("\nEnd: %Y-%m-%d %H:%M:%S %Z\n",
                                time.gmtime()))
     output.close()
@@ -641,6 +789,7 @@ def create_file(filename, contents):
 
 def main():
     setup_logging(logging.INFO, None)
+    signal(SIGHUP, sighup_handler)
 
     # For supporting multiple architectures and suites, we take command-line
     # argument(s) referring to section(s) in the configuration file.
@@ -658,15 +807,25 @@ def main():
                 for section_name in section_names]
 
     while True:
+        global got_sighup
         test_count = 0
-        precedence = None
 
-        for section in sorted(sections, key=lambda section: section.precedence()):
-            if precedence is None or section.precedence() <= precedence:
-                processed = section.run()
-                if processed > 0:
-                    test_count += processed
-                    precedence = section.precedence()
+        for section in sorted(sections, key=lambda section: (section.precedence(), section.sleep_until())):
+            test_count += section.run(do_processing=(test_count == 0 and not got_sighup))
+
+        if test_count == 0 and got_sighup:
+            # clear SIGHUP state after flushing all sections
+            got_sighup = False
+            continue
+
+        if test_count == 0:
+            # try to recycle old logs
+            # round robin recycling of all sections is ensured by the recycle_wait_until timestamps
+            idle_until = min([section.sleep_until() for section in sections])
+            for section in sorted(sections, key=lambda section: section.sleep_until(recycle=True)):
+                test_count += section.run(recycle=True)
+                if test_count > 0 and idle_until < time.time():
+                    break
 
         if test_count == 0:
             now = time.time()
